@@ -44,6 +44,20 @@ class DuplicatePair(Exception):
     """Found multiple trading pairs for the same naive lookup."""
 
 
+#: Data needed to identify a trading pair with human description.
+#:
+#: This is `(chain, exchange slug, base token, quote token)`.
+FeelessPair: TypeAlias = Tuple[ChainId, str, str, str]
+
+#: Data needed to identify a trading pair with human description.
+#:
+#: A version that can sepearate different fee variants.
+#: This is all Uniswap v3 pairs, as a single exchange
+#: supports the same pair with different pools having different fees.
+#:
+#: This is `(chain, exchange slug, base token, quote token, pool fee)`.
+FeePair: TypeAlias = Tuple[ChainId, str, str, str, float]
+
 
 #: Shorthand method to identify trading pairs when written down by a human.
 #:
@@ -54,7 +68,7 @@ class DuplicatePair(Exception):
 #: we usually refer to the "best" token which is the highest liquidty/volume trading
 #: pair on the particular exchange.
 #:
-HumanReadableTradingPairDescription: TypeAlias = Tuple[ChainId, str, str, str]
+HumanReadableTradingPairDescription: TypeAlias = FeePair | FeelessPair
 
 
 
@@ -451,19 +465,38 @@ class PandasPairUniverse:
 
     """
 
-    def __init__(self, df: pd.DataFrame, build_index=True):
+    def __init__(self,
+                 df: pd.DataFrame,
+                 build_index=True,
+                 exchange_universe: Optional[ExchangeUniverse]=None):
         """
-        :param df: The source DataFrame that contains all DEXPair entries
+        :param df:
+            The source DataFrame that contains all DEXPair entries
 
-        :param build_index: Build quick lookup index for pairs
+        :param build_index:
+            Build quick lookup index for pairs
+
+        :param exchange_universe:
+            Optional exchange universe needed for human-readable pair lookup.
+
+            We cannot properly resolve pairs unless we can map exchange names to their ids.
+            Currently optional, only needed by `get_pair()`.
         """
         assert isinstance(df, pd.DataFrame)
         self.df = df.set_index(df["pair_id"])
 
-        #: pair_id -> DEXPair mappings
+        #: pair_id -> raw dict data mappings
+        #:
+        #: Constructed in one pass from Pandas DataFrame.
+        #:
         #: Don't access directly, use :py:meth:`iterate_pairs`.
         self.pair_map: Dict[int, dict] = {}
 
+        #: pair_id -> constructed DEXPair cache
+        #:
+        #: Don't access directly, use :py:meth:`iterate_pairs`.
+        self.dex_pair_obj_cache: Dict[int, DEXPair] = {}
+
         # pair smart contract address -> DEXPair
         self.smart_contract_map = {}
 
@@ -474,10 +507,12 @@ class PandasPairUniverse:
         if build_index:
             self.build_index()
 
+        self.exchange_universe = exchange_universe
+
     def iterate_pairs(self) -> Iterable[DEXPair]:
         """Iterate over all pairs in this universe."""
-        for p in self.pair_map.values():
-            yield DEXPair.from_dict(p)
+        for pair_id in self.pair_map.keys():
+            yield self.get_pair_by_id(pair_id)
 
     def build_index(self):
         """Create pair_id -> data mapping.
@@ -497,11 +532,16 @@ class PandasPairUniverse:
         self.pair_map = self.df.T.to_dict()
         self.smart_contract_map = {d["address"].lower(): d for d in self.pair_map.values()}
 
-    def get_all_pair_ids(self) -> List[PrimaryKey]:
+    def get_all_pair_ids(self) -> Collection[PrimaryKey]:
+        """Get all pair ids in the data frame."""
         return self.df["pair_id"].unique()
 
-    def get_pair_ids_by_exchange(self, exchange_id: PrimaryKey):
-        """Get all pair ids on a specific exchange"""
+    def get_pair_ids_by_exchange(self, exchange_id: PrimaryKey) -> pd.DataFrame:
+        """Get all pair ids on a specific exchange.
+
+        :return:
+            Raw slide of DataFrame
+        """
         return self.df.loc[self.df["exchange_id"] == exchange_id]["pair_id"]
 
     def get_count(self) -> int:
@@ -511,6 +551,10 @@ class PandasPairUniverse:
     def get_pair_by_id(self, pair_id: PrimaryKey)  -> Optional[DEXPair]:
         """Look up pair information and return its data.
 
+        Uses a cached path. Constructing :py:class:`DEXPair`
+        objects is a bit slow, so this is a preferred method
+        if you need to access multiple pairs in a hot loop.
+
         :return:
             Nicely presented :py:class:`DEXPair`.
 
@@ -518,18 +562,17 @@ class PandasPairUniverse:
             return None.
         """
 
-        # First try the cached suing self.pair_map
+        # First try the cached paths
         if self.pair_map:
-            # Convert any pairs in-fly to DEXPair objects and store them.
-            # We do not initially construct these objects,
-            # as we do not know what pairs a strategy might access.
-            data = self.pair_map.get(pair_id)
 
-            if data is None:
-                return None
+            # First try object cache
+            obj = self.dex_pair_obj_cache.get(pair_id)
+            if not obj:
 
-            if isinstance(data, DEXPair):
-                return data
+                # Convert any pairs in-fly to DEXPair objects and store them.
+                # We do not initially construct these objects,
+                # as we do not know what pairs a strategy might access.
+                data = self.pair_map.get(pair_id)
 
             # Convert
             obj = DEXPair.from_dict(data)
@@ -566,7 +609,7 @@ class PandasPairUniverse:
         address = address.lower()
         assert self.smart_contract_map, "You need to build the index to use this function"
         data = self.smart_contract_map.get(address)
-        return DEXPair.from_dict(data)
+        return self.get_pair_by_id(data["pair_id"])
 
     def get_token(self, address: str) -> Optional[Token]:
         """Get a token that is part of any trade pair.
@@ -585,14 +628,15 @@ class PandasPairUniverse:
 
         token: Optional[Token] = None
 
+        assert len(self.pair_map) > 0, "This method can be only used with in-memory pair index"
+
         if address not in self.token_cache:
-            for pair_data in self.pair_map.values():
-                if pair_data["token0_address"] == address:
-                    p = DEXPair.from_dict(pair_data)
+            for pair_id in self.pair_map.keys():
+                p = self.get_pair_by_id(pair_id)
+                if p.token0_address == address:
                     token = Token(p.chain_id, p.token0_symbol, p.token0_address, p.token0_decimals)
                     break
-                elif pair_data["token1_address"] == address:
-                    p = DEXPair.from_dict(pair_data)
+                elif p.token1_address == address:
                     token = Token(p.chain_id, p.token1_symbol, p.token1_address, p.token1_decimals)
                     break
             self.token_cache[address] = token
@@ -608,8 +652,8 @@ class PandasPairUniverse:
 
         """
         tokens = set()
-        for pair_data in self.pair_map.values():
-            p = DEXPair.from_dict(pair_data)
+        for pair_id in self.pair_map.keys():
+            p = self.get_pair_by_id(pair_id)
             tokens.add(self.get_token(p.base_token_address))
             tokens.add(self.get_token(p.quote_token_address))
         return tokens
@@ -633,9 +677,10 @@ class PandasPairUniverse:
             a random pair is returned.g
 
         """
-        for pair in self.pair_map.values():
-            if pair["base_token_symbol"] == base_token_symbol and pair["quote_token_symbol"] == quote_token_symbol:
-                return DEXPair.from_dict(pair)
+        for pair_id in self.pair_map.keys():
+            pair = self.get_pair_by_id(pair_id)
+            if pair.base_token_symbol == base_token_symbol and pair.quote_token_symbol == quote_token_symbol:
+                return pair
         return None
 
     def get_one_pair_from_pandas_universe(self, exchange_id: PrimaryKey, base_token: str, quote_token: str, pick_by_highest_vol=False) -> Optional[DEXPair]:
@@ -696,6 +741,46 @@ class PandasPairUniverse:
 
         return None
 
+    def get_pair(self,
+                 chain_id: ChainId,
+                 exchange_slug: str,
+                 base_token: str,
+                 quote_token: str,
+                 fee_tier: Optional[float] = None
+                 ) -> DEXPair:
+        """Get a pair by its description.
+
+        The simplest way to access pairs in the pair universe.
+
+        To use this method, we must include `exchange_universe` in the :py:meth:`__init__`
+        as otherwise we do not have required look up tables.
+
+        :return:
+            The trading pair on the exchange.
+
+            Highest volume trading pair if multiple matches.
+
+        :raise NoPairFound:
+            In the case input data cannot be resolved.
+        """
+
+        assert isinstance(chain_id, ChainId)
+        assert type(exchange_slug) == str
+        assert type(base_token) == str
+        assert type(quote_token) == str
+
+        if fee_tier is not None:
+            assert type(quote_token) == float
+
+        assert self.exchange_universe is not None, "You need to set exchange_universe argument in the construction to use this method"
+
+        if fee_tier:
+            desc = (chain_id, exchange_slug, base_token, quote_token, fee_tier)
+        else:
+            desc = (chain_id, exchange_slug, base_token, quote_token,)
+
+        return self.get_pair_by_human_description(self.exchange_universe, desc)
+
     def get_pair_by_human_description(self,
                                       exchange_universe: ExchangeUniverse,
                                       desc: HumanReadableTradingPairDescription) -> DEXPair:
@@ -1095,10 +1180,11 @@ def filter_for_stablecoins(pairs: pd.DataFrame, mode: StablecoinFilteringMode) -
     return our_pairs
 
 def resolve_pairs_based_on_ticker(
-        df: pd.DataFrame,
-        chain_id: ChainId,
-        exchange_slug: str,
-        pair_tickers: Set[Tuple[str, str]],
+    df: pd.DataFrame,
+    chain_id: ChainId,
+    exchange_slug: str,
+    pair_tickers: set[tuple[str, str] | tuple[str, str, BasisPoint]],
+    sorting_criteria: str = "buy_volume_all_time",
 ) -> pd.DataFrame:
     """Resolve symbolic trading pairs to their internal integer primary key ids.
 
@@ -1164,9 +1250,18 @@ def resolve_pairs_based_on_ticker(
     # Create list of conditions to filter out dataframe,
     # one condition per pair
     conditions = []
-    for base, quote in pair_tickers:
-        condition = (df["base_token_symbol"].str.lower() == base.lower()) & \
-                    (df["quote_token_symbol"].str.lower() == quote.lower())
+    for base, quote, *fee in pair_tickers:
+        condition = (
+            (df["base_token_symbol"].str.lower() == base.lower())
+            & (df["quote_token_symbol"].str.lower() == quote.lower())
+            & (df["exchange_slug"].str.lower() == exchange_slug.lower())
+            & (df["chain_id"] == chain_id.value)
+        )
+
+        # also filter by pair fee if pair ticker specifies it
+        if len(fee) > 0:
+            condition &= (df["fee"] == fee[0])
+
         conditions.append(condition)
 
     # OR call conditions together
@@ -1175,19 +1270,20 @@ def resolve_pairs_based_on_ticker(
 
     # Sort by the buy volume as a preparation
     # for the scam filter
-    df_matches = df_matches.sort_values(by="buy_volume_all_time", ascending=False)
+    df_matches = df_matches.sort_values(by=sorting_criteria, ascending=False)
 
     result_pair_ids = set()
 
     # Scam filter.
     # Pick the tokens by the highest buy volume to the result map.
-    for test_pair in pair_tickers:
-        for idx, row in df_matches.iterrows():
-            if row["base_token_symbol"] == test_pair[0] and \
-               row["quote_token_symbol"] == test_pair[1] and \
-               row["exchange_slug"] == exchange_slug:
-                    result_pair_ids.add(row["pair_id"])
-                    break
+    for base, quote, *_ in pair_tickers:
+        for _, row in df_matches.iterrows():
+            if (
+                row["base_token_symbol"] == base
+                and row["quote_token_symbol"] == quote
+            ):
+                result_pair_ids.add(row["pair_id"])
+                break
 
     result_df = df.loc[df["pair_id"].isin(result_pair_ids)]
     return result_df
