@@ -1,6 +1,7 @@
 """Helpers to create Pandas dataframes for per-pair analytics."""
 
 import logging
+import warnings
 from typing import Optional, Tuple, Iterable
 
 import pandas as pd
@@ -90,6 +91,35 @@ class PairGroupedUniverse:
             raise KeyError(f"No OHLC samples for pair id {pair_id} in {self}") from e
         return pair
 
+    def get_last_entries_by_pair_and_timestamp(self,
+            pair_id: PrimaryKey,
+            timestamp: pd.Timestamp,
+            small_time=pd.Timedelta(seconds=1),
+        ) -> pd.DataFrame:
+        """Get samples for a single pair before a timestamp.
+
+        Return a DataFrame slice containing all datapoints before the timestamp.
+
+        :param pair_id:
+            Integer id for a trading pair
+
+        :param timestamp:
+            Get all samples excluding this timestamp.
+
+        :return:
+            Dataframe that contains samples for a single trading pair.
+
+            Indexed by timestamp.
+
+        :raise KeyError:
+            If we do not have data for pair_id
+        """
+        pair_candles = self.get_samples_by_pair(pair_id)
+        # Watch out for inclusive timestamp
+        # https://stackoverflow.com/questions/49962417/why-does-loc-have-inclusive-behavior-for-slices
+        adjusted_timestamp = timestamp - small_time
+        return pair_candles.loc[:adjusted_timestamp]
+
     def get_all_pairs(self) -> Iterable[Tuple[PrimaryKey, pd.DataFrame]]:
         """Go through all liquidity samples, one DataFrame per trading pair."""
         for pair_id, data in self.pairs:
@@ -97,8 +127,15 @@ class PairGroupedUniverse:
 
     def get_pair_ids(self) -> Iterable[PrimaryKey]:
         """Get all pairs present in the dataset"""
-        for pair_id, data in self.pairs:
-            yield pair_id
+        with warnings.catch_warnings():
+            # FutureWarning: In a future version of pandas, a length 1 tuple will be returned when 
+            # iterating over a groupby with a grouper equal to a list of length 1. 
+            # Don't supply a list with a single grouper to avoid this warning.
+            
+            warnings.simplefilter("ignore")
+        
+            for pair_id, data in self.pairs:
+                yield pair_id
 
     def get_all_samples_by_timestamp(self, ts: pd.Timestamp) -> pd.DataFrame:
         """Get list of candles/samples for all pairs at a certain timepoint.
